package labss.lang.il.visitors;

import java.util.HashMap;
import java.util.Map;

import labss.lang.il.model.*;
import labss.lang.il.model.impl.*;

public class InterpreterVisitor implements ILVisitor {
	protected Map<String, Literal> bindings;
	protected Literal result = null;

	public static Literal interpret(ILEntity program) {
		return interpret(program, new HashMap<String, Literal>());
	}
	public static Literal interpret(ILEntity program, Map<String, Literal> bindings) {
		InterpreterVisitor v = new InterpreterVisitor(bindings);
		program.accept(v);
		return v.result;
	}

	protected InterpreterVisitor(Map<String, Literal> bindings) {
		this.bindings = bindings;
	}

	protected final void execute(Statement stm) {
		stm.accept(this);
	}
	protected final Literal evaluate(Expression exp) {
		exp.accept(this);
		return result;
	}
	protected final boolean booleanValue(Expression exp) {
		return evaluate(exp).booleanValue();
	}
	protected final int intValue(Expression exp) {
		return evaluate(exp).intValue();
	}
	protected final void setResult(boolean value) {
		result = ILConcreteFactory.instance().createBoolean(value);
	}
	protected final void setResult(int value) {
		result = ILConcreteFactory.instance().createInt(value);
	}

	public void visit(Empty stm) {
	}

	public void visit(While stm) {
		while (booleanValue(stm.getCondition()))
			execute(stm.getTrueBody());
	}

	public void visit(If stm) {
		if (booleanValue(stm.getCondition()))
			execute(stm.getTrueBody());
	}
	public void visit(IfElse stm) {
		if (booleanValue(stm.getCondition()))
			execute(stm.getTrueBody());
		else
			execute(stm.getFalseBody());
	}

	public void visit(Assignment stm) {
		bindings.put(stm.getVariable().getName(), evaluate(stm.getExpression()));
	}

	public void visit(Block stm) {
		for (Statement s : stm.getStatements())
			s.accept(this);
	}

	public void visit(Bool exp) {
		setResult(exp.booleanValue());
	}
	public void visit(Int exp) {
		setResult(exp.intValue());
	}

	public void visit(Variable exp) {
		result = bindings.get(exp.getName());
		if (result == null)
			throw new IllegalStateException("Variabile non inizializzata: "+exp.getName());
	}


	public void visit(Addition exp) {
		setResult(intValue(exp.getExp1()) + intValue(exp.getExp2()));
	}
	public void visit(Subtraction exp) {
		setResult(intValue(exp.getExp1()) - intValue(exp.getExp2()));
	}
	public void visit(Multiplication exp) {
		setResult(intValue(exp.getExp1()) * intValue(exp.getExp2()));
	}
	public void visit(Division exp) {
		setResult(intValue(exp.getExp1()) / intValue(exp.getExp2()));
	}

	public void visit(Equals exp) {
		setResult(intValue(exp.getExp1()) == intValue(exp.getExp2()));
	}
	public void visit(GreaterThan exp) {
		setResult(intValue(exp.getExp1()) > intValue(exp.getExp2()));
	}
	public void visit(LessThan exp) {
		setResult(intValue(exp.getExp1()) < intValue(exp.getExp2()));
	}

	public void visit(And exp) {
		setResult(booleanValue(exp.getExp1()) && booleanValue(exp.getExp2()));
	}
	public void visit(Or exp) {
		setResult(booleanValue(exp.getExp1()) || booleanValue(exp.getExp2()));
	}
	public void visit(Not exp) {
		setResult(!booleanValue(exp.getExp()));
	}
}
