package labss.lang.il.visitors;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;

import labss.lang.il.model.Expression;
import labss.lang.il.model.ILEntity;
import labss.lang.il.model.impl.Addition;
import labss.lang.il.model.impl.And;
import labss.lang.il.model.impl.Assignment;
import labss.lang.il.model.impl.Bool;
import labss.lang.il.model.impl.Division;
import labss.lang.il.model.impl.Empty;
import labss.lang.il.model.impl.Equals;
import labss.lang.il.model.impl.GreaterThan;
import labss.lang.il.model.impl.If;
import labss.lang.il.model.impl.IfElse;
import labss.lang.il.model.impl.Int;
import labss.lang.il.model.impl.LessThan;
import labss.lang.il.model.impl.Multiplication;
import labss.lang.il.model.impl.Not;
import labss.lang.il.model.impl.Or;
import labss.lang.il.model.impl.Subtraction;
import labss.lang.il.model.impl.Variable;
import labss.lang.il.model.impl.While;

public class ValidatorVisitor extends ILTraverseAllVisitor {
	public enum DataType { BOOL, INT, UNKNOWN };
	private Map<String, DataType> declaredVariables = new HashMap<String, DataType>();
	private DataType dataType = DataType.UNKNOWN;
	private int lineNumber = 0;
	private PrintStream log = System.out;
	private int errors=0, warnings=0;

	public static int validate(ILEntity program) {
		ValidatorVisitor visitor = new ValidatorVisitor();
		program.accept(visitor);
		return visitor.errors;
	}

	public void incLineNumber() {
		lineNumber++;
	}
	public void addWarning(String message) {
		warnings++;
		log.println("Warning at line "+lineNumber+". "+message);
	}
	public void addError(String message) {
		errors++;
		log.println("Error at line "+lineNumber+". "+message);
	}

	protected DataType getDataType(String var) {
		return declaredVariables.containsKey(var) ? declaredVariables.get(var) : DataType.UNKNOWN;
	}
	protected void ensureDataType(String var) {
		DataType dataType = getDataType(var);
		if (this.dataType == DataType.UNKNOWN)
			this.dataType = dataType;
		else if (dataType == DataType.UNKNOWN)
			dataType = this.dataType;
		else if (this.dataType != dataType)
			addError("Incompatible type: expected "+this.dataType+" found "+dataType);
	}
	protected DataType getDataType() {
		return dataType;
	}
	protected void setDataType(DataType dataType) {
		this.dataType = dataType;
	}
	protected void setDataType(String var) {
		if (!declaredVariables.containsKey(var) || getDataType(var).equals(DataType.UNKNOWN))
			declaredVariables.put(var, getDataType());
		else
			setDataType(getDataType(var));
	}
	protected DataType assertDataType(DataType dataType) {
		DataType oldDataType = this.dataType;
		if (this.dataType == DataType.UNKNOWN)
			return this.dataType = dataType;
		else if (this.dataType != dataType) {
			addError("Incompatible type: expected "+this.dataType+" found "+dataType);
			this.dataType = dataType;
		}
		return oldDataType;
	}

	
	protected void ensureDefined(String var) {
		if (!declaredVariables.containsKey(var))
			declaredVariables.put(var, DataType.UNKNOWN);
	}
	protected void assertDefined(String var) {
		if (!declaredVariables.containsKey(var) || getDataType(var).equals(DataType.UNKNOWN)) {
			addWarning("Expected parameter: "+var);
			declaredVariables.put(var, getDataType());
		}
	}

	protected void assertBoolean(Expression exp) {
		setDataType(DataType.BOOL);
		exp.accept(visitor1);
	}

	@Override
	public void visit(If entity) {
		incLineNumber();
		assertBoolean(entity.getCondition());
		entity.getTrueBody().accept(visitor1);
	}

	@Override
	public void visit(IfElse entity) {
		incLineNumber();
		assertBoolean(entity.getCondition());
		entity.getTrueBody().accept(visitor1);
		entity.getFalseBody().accept(visitor1);
	}

	@Override
	public void visit(While entity) {
		incLineNumber();
		assertBoolean(entity.getCondition());
		entity.getTrueBody().accept(visitor1);
	}

	@Override
	public void visit(Assignment entity) {
		incLineNumber();
		String varName = entity.getVariable().getName();
		ensureDefined(varName);
		setDataType(getDataType(varName));
		entity.getExpression().accept(visitor1);
		setDataType(varName);
	}

	@Override
	public void visit(Empty stm) {
		incLineNumber();
		super.visit(stm);
	}

	@Override
	public void visit(Variable exp) {
		String varName = exp.getName();
		assertDefined(varName);
		setDataType(assertDataType(getDataType(varName)));
	}
		
	@Override
	public void visit(Bool exp) {
		setDataType(assertDataType(DataType.BOOL));
	}
	@Override
	public void visit(Int exp) {
		setDataType(assertDataType(DataType.INT));
	}

	@Override
	public void visit(Addition exp) {
		DataType oldDataType = assertDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Subtraction exp) {
		DataType oldDataType = assertDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Multiplication exp) {
		DataType oldDataType = assertDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Division exp) {
		DataType oldDataType = assertDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Equals exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		setDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(GreaterThan exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		setDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(LessThan exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		setDataType(DataType.INT);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(And exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Or exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		super.visit(exp);
		setDataType(oldDataType);
	}
	@Override
	public void visit(Not exp) {
		DataType oldDataType = assertDataType(DataType.BOOL);
		super.visit(exp);
		setDataType(oldDataType);
	}
}
